@c -*-texinfo-*-

@node Introduction, Uniformity Testing, Top, Top
@chapter Introduction

This report presents a library for comparing the distribution of
multidimensional data against a uniform distribution, using the minimum
spanning tree-based approach described in @bibref{Smith 1984}.  The
complete source code for the library is included, along with discussion
of the rationale behind its structure.  Finally, its performance is
analyzed, and possibilities for further refinements are noted.

The audience for this document is twofold.  First, it is intended as a
class semester report.  For this purpose, the first part of the next
chapter (@pxref{Uniformity Testing}) and the final chapter
(@pxref{Discussion}) are probably the most interesting parts.  A
secondary purpose for this document is to present the source code for
the library as a literate program, as described below.  For this reason,
some algorithms, such as those for maintaining a binary heap, are
described in more depth than would otherwise be necessary.

@menu
* How to Read a TexiWEB::       
* Coding Style::                
* License::                     
@end menu

@node How to Read a TexiWEB, Coding Style, Introduction, Introduction
@section How to Read a @TexiWEB{}

The code in this report is written in ANSI/ISO C89 using @TexiWEB{}, a
literate programming system.  Literate programming is a philosophy
that regards software as a form of literature.  The ideas behind
literate programming have been around for a long time, but the term
itself was invented by famous computer scientist Donald Knuth in 1984,
who wrote two of his most famous programs (@TeX{} and @MF{}) with a
literate programming system of his own design.  That system, called
@WEB{}, inspired the form and much of the syntax of @TexiWEB{}.

A @TexiWEB{} document is a C program that has been cut into sections,
rearranged, and annotated, with the goal to make the program as a whole
as comprehensible as possible to a reader who starts at the beginning
and reads the entire program in order.  To this end, each section of a
@TexiWEB{} program is assigned both a number and a name.  Section
numbers are assigned sequentially, starting from 1 with the first
section, and they are used for cross-references between sections.
Section names are English words or phrases assigned by the @TexiWEB{}
program's author to describe the function of the corresponding code.

Here's a sample @TexiWEB{} section:

@begincode{}@tabalign{}@b{19.} @value{LANG}@anchor{Clear hash table entries}Clear hash table entries @little{19}@value{RANG} @value{IS}@cr
@tabalign{}@b{for} (@cleartabs{}@wtab{}@i{i} = 0; @i{i} @math{<} @i{hash}@value{RARR}@i{m}; @i{i}@math{+}@math{+}) @cr
  @tabalign{}@IND{2em}@i{hash}@value{RARR}@i{entry}[@cleartabs{}@wtab{}@i{i}] = @t{NULL};@cr
@endcode{}

You should notice how the first line gives the section's number,
followed by its name and again its number within angle brackets.  
@iftex
If you know C, you'll also notice that the C operator -> has been
replaced by the nicer-looking arrow |->|.  @TexiWEB{} makes an attempt
to ``prettify'' C in a few ways like this.  The table below lists most
of these substitutions:

@multitable @columnfractions .25 .15 .18 .15
@item @tab -> @tab becomes @tab |->|
@item @tab 0x12ab @tab becomes @tab |0x12ab|
@item @tab 0377 @tab becomes @tab |0377|
@item @tab 1.2e34 @tab becomes @tab |1.2e34|
@end multitable

In addition,| + |and| - |are written as superscripts when used to
indicate sign, as in |-5| and |+10|.
@end iftex

@ifnotinfo
In @TexiWEB{}, C's reserved words are shown like this: |int|, |struct|,
|while|, and so on.  Types defined with |typedef| or with |struct|,
|union|, and |enum| tags are also shown this way.  Identifiers in all
capital letters (often names of macros) are shown like this: |BUFSIZ|,
|EOF|, |ERANGE|, and so on.  Other identifiers are shown like this:
|getc|, |argv|, |strlen|, and so on.
@end ifnotinfo

@iftex
Sometimes it is desirable to talk about mathematical expressions, rather
than C expressions.  When this is done, mathematical operators (@<=,
@>=) instead of C operators (|<=|, |>=|) are used.  In particular,
equality is indicated with @= instead of = in order to minimize
potential confusion.
@end iftex

Code segments often contain references to other code segments, shown as
a section name and number within angle brackets.  These act something
like macros, in that they stand for the corresponding replacement text.
For instance, consider the following segment:

@begincode{}@tabalign{}@b{15.} @value{LANG}@anchor{Initialize hash table}Initialize hash table @little{15}@value{RANG} @value{IS}@cr
@tabalign{}@i{hash}@value{RARR}@i{m} = 13;@cr
@tabalign{}@refcode{Clear hash table entries,Clear hash table entries,19}@cr
@endcode{}

This means that the code for `Clear hash table entries' should be
inserted as part of `Initialize hash table'.  Since the name of a
section explains what it does, it's often unnecessary to know anything
more.  But if you do want more detail, the section number @little{19} in
@refcode{Clear hash table entries,Clear hash table entries,19} can
easily be used to find the full text and annotations for `Clear hash
table entries'.  In addition, at the bottom of section 19 you will find
a note reading `@little{This code is included in segment
@refalso{Initialize hash table, , 15}.}', making it easy to move back to
section 15 that includes it.

It is possible to have multiple sections with the same name.  When a
name that corresponds to multiple sections is referenced, code from all
the sections with that name is substituted, in order of appearance.  For
instance, the following shows another line of code in `Initialize hash
table':

@begincode{}@tabalign{}@b{16.} @value{LANG}@anchor{Initialize hash table-2}Initialize hash table @little{15}@value{RANG} @math{+}@value{IS}@cr
@tabalign{}@i{hash}@value{RARR}@i{n} = 0;@cr
@endcode{}

When multiple sections have the same name, those sections end with a
note listing the numbers of all other same-named sections.  In this
case, the note would read `@little{See also segment @refalso{Initialize
hash table, , 15}.}'.

@ignore
Occasionally a section of code may be both irrelevant to the text and
uninteresting to read.  When this happens, it is omitted.  References to
such a section are marked as @refcode{Section name, Node name, OMITTED}.

@end ignore
When a @TexiWEB{} program is converted to C, conversion conceptually
begins from sections whose names are file names; e.g.,
@value{LANG}@file{foo.c} @little{37}@value{RANG}.  Within these
sections, all of the section references are expanded, then any
references within those sections are expanded, and so on.  When
expansion is complete, the specified files are written out.  

Finally, another useful resource in reading a @TexiWEB{} is the index,
which contains an entry for the points of declaration of every section
name, function, type, structure, union, global variable, and macro.
Declarations within functions are not indexed.

(This section is adapted from ``Reading the Code,'' @bibref{Pfaff
2000}.  It was inspired by ``How to read a @WEB{},'' @bibref{Knuth
1992}.)

@node Coding Style, License, How to Read a TexiWEB, Introduction
@section Coding Style

The source code here complies to the requirements imposed by ANSI/ISO
C89 and ISO/IEC C99.  Only features present in C89 are used, but the
code is forward-compatible with C99.

Most of the GNU Coding Standards are followed.  Indentation style is an
exception: in print, to save vertical space, K&R indentation style is
used instead of GNU style.  The code here also conforms to Ben Pfaff's
published personal coding standards.

@references
@bibref{ISO 1990};
@bibref{ISO 1999};
@bibref{FSF 2001}, section ``Writing C'';
@bibref{Pfaff 2001}.

@node License,  , Coding Style, Introduction
@section License

The library code itself is under a liberal BSD-style license:

@<BSD License@> =
/* Copyright (c) 2001 Ben Pfaff <pfaffben@@msu.edu>.  All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:

     - Redistributions of source code must retain the above copyright
       notice, this list of conditions, and the following disclaimer.

     - Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions, and the following disclaimer in
       the documentation and/or other materials provided with the
       distribution.

   There is NO WARRANTY, not even for MERCHANTABILITY or FITNESS FOR A
   PARTICULAR PURPOSE. */
@

@noindent 
TexiWEB and some of the library's test code is derived from GNU libavl.
This code falls under the GNU General Public License:

@<GNU GPL@> =
/* Copyright (C) 2001 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The author may be contacted at <pfaffben@@msu.edu> on the Internet,
   or as Ben Pfaff, 12167 Airport Rd, DeWitt MI 48820, USA through
   more mundane means. */
@
