@c -*-texinfo-*-

@node Priority Queues, Other Classes, Minimum Spanning Tree Computation, Top
@chapter Priority Queues

Prim's algorithm for computing a minimum spanning tree requires the use
of a priority queue to determine the order of insertion of edges into
the tree.  This chapter presents two different priority queues for this
purpose.  The first is based on simple binary heaps, the second on more
complicated and asymptotically faster Fibonacci heaps.

@menu
* PQ Class::                    
* Binary Heap PQ::              
* Fibonacci Heap PQ::           
@end menu

@node PQ Class, Binary Heap PQ, Priority Queues, Priority Queues
@section Class Structure

In order to make it easy to switch among implementatations, the priority
queue is defined in @(pq.h@> as a ``class'' structure of function
pointers.  Notice that because @(pq.h@> is not intended to be included
by programs using the uniformity-testing library, only by
@(mst-prim.c@>, it does not restrict itself to the |unf_| namespace:

@(pq.h@> =
@<BSD License@>

#ifndef PQ_H
#define PQ_H 1

struct unf_mem;

struct pq_class @
  {@-
    struct pq *(*create) (struct unf_mem *mem, int n);
    void (*discard) (struct pq *pq);
    int (*extract_min) (struct pq *pq);
    void (*decrease_key) (struct pq *pq, int vertex, double key);
    int (*count) (const struct pq *pq);
    double (*key) (const struct pq *pq, int vertex);
    void (*verify) (const struct pq *pq);
    void (*dump) (const struct pq *pq);
  };@+

#endif /* pq.h */
@

The abstract model behind |struct pq_class| is a first-in-smallest-out
priority queue.  The items in the queue have two parts: an integer
``vertex'' number between 0 and |n - 1|, where |n| is the maximum
capacity of the queue, and a ``key'' stored as a floating-point number.
The items are ordered based on the key field.  An instance of |struct
pq_class| must implement these abstract operations:

@table @asis
@item |struct pq *create (struct unf_mem *mem, int n);|
Creates and returns a new priority queue containing |n| items initially,
all of which initially have key @altmath{+\infty, +oo}.  Any needed
memory allocation for the priority queue must be done with the functions
in |mem|.  Returns a null pointer if the priority queue cannot be
created.

@item |void discard (struct pq *pq);|
Destroys priority queue |pq| and frees any associated memory.

@item |int extract_min (struct pq *pq);|
Deletes the item within |pq| having minimum key and returns its vertex.
If |pq| is empty, behavior is undefined.

@item |void decrease_key (struct pq *pq, int vertex, double key);|
Finds the item in |pq| with vertex |vertex| and changes its key to
|key|.  Behavior is undefined if |pq| does not contain an item with
vertex |vertex| or if |key| is greater than that item's current key.

@item |int count (const struct pq *pq);|
Returns the number of items remaining in |pq|.  This will always be less
than or equal to |n| passed to |create()|.

@item |double key (const struct pq *pq, int vertex);|
Returns the key of the item in |pq| with vertex |vertex|.  Returns
|-DBL_MAX| if there is no vertex |vertex| in |pq|.  Behavior is
undefined if |vertex| is not in the valid range.

@item |void verify (const struct pq *pq);|
(For debugging purposes.)  Verifies that |pq|'s state is internally
consistent.

@item |void dump (const struct pq *pq);|
(For debugging purposes.)  Prints a human-readable representation of
|pq|'s internal state to |stdout|.
@end table

@node Binary Heap PQ, Fibonacci Heap PQ, PQ Class, Priority Queues
@section Binary Heap

File @(pq-bin-heap.c@> implements binary heaps.  It begins like this:

@(pq-bin-heap.c@> =
@<BSD License@>

#include <assert.h>
#include <float.h>
#include <stdio.h>
#include "pq.h"
#include "uniformity.h"

@

The class of binary heaps is a subset of the class of binary trees.  A
binary tree is a binary heap if:

@enumerate 1
@item
For every node except the root, the key value in its parent node is less
or equal than its own key value (the ``heap property''); @emph{and}

@item
The binary tree is ``nearly complete,'' meaning that every level in the
tree is filled, except possibly for the bottommost, which must be filled
in left-to-right order without gaps.
@end enumerate

It is not necessary, but here we will also assume that key values in
heap nodes are distinct.  Here are two examples of heaps, with the key
values shown inside the node circles:

@center @image{b-heap-ex}

A heap can be conveniently represented as an array, with no need for
pointers, by writing down its elements from left to right, top to
bottom.  The two heaps pictured above would have the following
representations as arrays:

@display
(2, 4, 5, 7, 11, 6, 8, 9, 10, 12)
(1, 2, 6, 3, 5, 7)
@end display

@noindent 
The first element of the array, at index 1, is the heap's root.  To get
from the |i|'th element to its left child, simply multiply |i| by 2; to
get to its right child, multiply by 2 and add 1; to get to its parent,
divide by 2 and discard any remainder.  If the result is between 1 and
the number of nodes in the heap, then it is the index of the desired
node; otherwise, there is no such node in the heap.  We can easily
implement these ``movement'' operations in C:

@(pq-bin-heap.c@> +=
/* Heap movement operations. */
#define parent(i) ((i) / 2)     /* Get node |i|'s parent. */
#define left(i) ((i) * 2)       /* Get node |i|'s left child. */
#define right(i) (left (i) + 1) /* Get node |i|'s right child. */

@

Our binary heap priority queue structure based on these ideas looks like
this:

@(pq-bin-heap.c@> +=
/* Binary heap priority queue. */
struct pq @
  {@-
    struct unf_mem *mem; /* Memory allocator. */
    int n;		/* Number of items in tree. */
    int m;		/* Maximum number of items in tree. */

    /* Heap contents. */
    double *key;	/* Keys. */
    int *vertex;	/* Satellite information (vertex number). */

    /* Reverse index. */
    int *index;		/* Heap index by vertex number; 0 = not present. */
  };@+

@

@noindent
Most of this structure should be self-explanatory.  Member |key| points
to the heap itself; |vertex| is a parallel array that contains the
companion vertex values.  Both of these arrays have an extra unused
element at the beginning, which allows for 1-based indexing and
simplifies the movement macros above.@footnote{On many architectures some
pointer trickery could be used instead, but this is technically
undefined according to @bibref{ISO 1990}, section 6.3.6.  See
@bibref{Summit 1999}, question 6.17, for more information.}  Member
|index| is used to store the indexes into |key| and |vertex| that
contain particular vertex values.  This array is necessary for fast
lookup of particular vertices, since there is no quick way to search a
heap.

The following sections implement the abstract operations from |struct
pq_class| on binary heaps.

@references
@bibref{Cormen 1990}, section 7.1;
@bibref{Knuth 1973}, section 5.2.3.

@menu
* Binary heap creation::        
* Binary heap destruction::     
* Binary heap extraction::      
* Binary heap decrease key::    
* Binary heap count::           
* Binary heap key lookup::      
* Binary heap verification::    
* Binary heap dump::            
* Binary heap class implementation::  
@end menu

@node Binary heap creation, Binary heap destruction, Binary Heap PQ, Binary Heap PQ
@subsection Creation

Creation and initialization of the heap is straightforward:

@(pq-bin-heap.c@> +=
static void pq_discard (struct pq *pq);

static struct pq *@
pq_create (struct unf_mem *mem, int n) @
{
  struct pq *pq;
  int i;

  /* Allocate memory for heap itself. */
  assert (n >= 0);
  pq = mem->unf_alloc (sizeof *pq);
  if (pq == NULL)
    return NULL;

  /* Allocate memory for heap's members. */
  pq->mem = mem;
  pq->key = mem->unf_alloc (sizeof *pq->key * (n + 1));
  pq->vertex = mem->unf_alloc (sizeof *pq->vertex * (n + 1));
  pq->index = mem->unf_alloc (sizeof *pq->index * (n + 1));
  if (pq->key == NULL || pq->vertex == NULL || pq->index == NULL) @
    {@-
      pq_discard (pq);
      return NULL;
    }@+

  /* Initialize heap. */
  pq->n = pq->m = n;
  for (i = 1; i <= n; i++) @
    {@-
      pq->key[i] = DBL_MAX;
      pq->vertex[i] = i;
      pq->index[i] = i;
    }@+

  return pq;
}

@

@node Binary heap destruction, Binary heap extraction, Binary heap creation, Binary Heap PQ
@subsection Destruction

Destruction simply discards the memory that was allocated at creation
time:

@(pq-bin-heap.c@> +=
static void @
pq_discard (struct pq *pq) @
{
  assert (pq != NULL);

  pq->mem->unf_free (pq->key);
  pq->mem->unf_free (pq->vertex);
  pq->mem->unf_free (pq->index);
  pq->mem->unf_free (pq);
}

@

@node Binary heap extraction, Binary heap decrease key, Binary heap destruction, Binary Heap PQ
@subsection Minimum Value Extraction

Extracting the minimum value from a binary heap can be divided into
three steps.

@(pq-bin-heap.c@> +=
static int
pq_extract_min (struct pq *pq)
{
  int min;
  
  assert (pq != NULL && pq->n > 0);

  @<Save minimum value of binary heap@>
  @<Delete root from binary heap@>
  @<Restore binary heap property@>

  return min - 1;
}
@

The minimum value in a binary heap is the root, so finding the minimum
value is easy:

@<Save minimum value of binary heap@> =
min = pq->vertex[1];
@

The interesting part is deleting the root node and rearranging the
resultant binary tree so that it is still a heap.  The successful
approach is to move the last (|n|'th) node in the heap into the root
position, then move down the tree swapping nodes until the heap property
is satisfied.  Here's an example of the effects of heap deletion, with
the original heap at the left and the heap after at the right and
the intermediate steps in between:

@center @image{b-heap-del}

We start by deleting the minimum node from the reverse index, moving the
last node to the root, and decrementing the tree's node count:

@<Delete root from binary heap@> =
assert (min >= 1 && min <= pq->m);
pq->index[min] = 0;

pq->vertex[1] = pq->vertex[pq->n];
pq->key[1] = pq->key[pq->n];
pq->index[pq->vertex[1]] = 1;

pq->n--;
  
@

To restore the heap property, we start from the root and move down the
tree, swapping the current node with the smallest of its children.  If
the current node is smaller than its children, we can stop:

@<Restore binary heap property@> =
{
  int i = 1;

  for (;;) @
    {@-
      /* Find index of smallest of |i| or its children as |smallest|. */
      int l = left (i);
      int r = right (i);
      int smallest = i;
      if (l <= pq->n && pq->key[l] < pq->key[smallest])
        smallest = l;
      if (r <= pq->n && pq->key[r] < pq->key[smallest])
        smallest = r;
      if (smallest == i)
        break;

      /* Swap nodes |i| and |smallest|. */
      {
        int t_vertex = pq->vertex[i];
        double t_key = pq->key[i];
        pq->vertex[i] = pq->vertex[smallest];
        pq->key[i] = pq->key[smallest];
        pq->vertex[smallest] = t_vertex;
        pq->key[smallest] = t_key;
        pq->index[pq->vertex[i]] = i;
        pq->index[pq->vertex[smallest]] = smallest;
      }

      i = smallest;
    }@+
}

@

@references
@bibref{Cormen 1990}, section 7.5;
@bibref{Knuth 1973}, section 5.2.3.

@node Binary heap decrease key, Binary heap count, Binary heap extraction, Binary Heap PQ
@subsection Decreasing Keys

To decrease the key value of a heap item, we cause it to ``bubble up''
the tree until its parent has a smaller value than it does or it reaches
the root.  We can always do this because we replace smaller values by
larger values, so the proper relationship between the keys replaced and
their children's keys is maintained.  In the example below, the key of
the node labeled 5 is decreased to 0:

@center @image{b-heap-dec}

@noindent
No further discussion should be necessary:

@(pq-bin-heap.c@> +=
static void @
pq_decrease_key (struct pq *pq, int vertex, double key) @
{
  int i, p;

  assert (pq != NULL);
  assert (vertex >= 0 && vertex < pq->m);
  assert (pq->index[vertex + 1] != 0);
  assert (key <= pq->key[pq->index[vertex + 1]]);

  for (i = pq->index[vertex + 1]; i > 1; i = p) @
    {@-
      p = parent (i);
      if (pq->key[p] < key)
	break;

      pq->key[i] = pq->key[p];
      pq->vertex[i] = pq->vertex[p];
      pq->index[pq->vertex[i]] = i;
    }@+

  pq->key[i] = key;
  pq->vertex[i] = vertex + 1;
  pq->index[pq->vertex[i]] = i;
}

@

@references
@bibref{Cormen 1990}, section 7.5, exercise 4.

@node Binary heap count, Binary heap key lookup, Binary heap decrease key, Binary Heap PQ
@subsection Count

The count function just returns |struct pq|'s |n| value:

@(pq-bin-heap.c@> +=
static int @
pq_count (const struct pq *pq) @
{
  assert (pq != NULL);

  return pq->n;
}

@

@node Binary heap key lookup, Binary heap verification, Binary heap count, Binary Heap PQ
@subsection Key Lookup

The |index| array is designed for key lookup, making this function easy:

@(pq-bin-heap.c@> +=
static double @
pq_key (const struct pq *pq, int vertex) @
{
  int i;

  assert (pq != NULL && vertex >= 0 && vertex < pq->m);

  i = pq->index[vertex + 1];
  return i != 0 ? pq->key[i] : -DBL_MAX;
}

@

@node Binary heap verification, Binary heap dump, Binary heap key lookup, Binary Heap PQ
@subsection Verification

The verification function checks that the heap property is satisfied and
that the |index| array is properly set up:

@(pq-bin-heap.c@> +=
static void @
pq_verify (const struct pq *pq) @
{
  int i;
  int error = 0;
  
  assert (pq != NULL);
  assert (pq->m > 0);
  assert (pq->n <= pq->m);

  for (i = 2; i <= pq->n; i++) @
    {@-
      int parent = parent (i);
      if (pq->key[parent] > pq->key[i]) @
	{@-
	  printf ("Heap property for %d:%d violated: %g > %g\n",
		  parent, i, pq->key[parent], pq->key[i]);
	  error = 1;
	}@+
    }@+

  for (i = 1; i <= pq->n; i++)
    if (pq->index[i] != 0 && pq->vertex[pq->index[i]] != i) @
      {@-
	printf ("index for %d points to position %d, which contains %d\n",
		i, pq->index[i], pq->vertex[pq->index[i]]);
	error = 1;
      }@+

  assert (error == 0);
}

@

@node Binary heap dump, Binary heap class implementation, Binary heap verification, Binary Heap PQ
@subsection Dumping

The dump function prints out an array representation of the binary heap
to |stdout|.  Vertex numbers and (finite) keys are printed separated by
colons.

@(pq-bin-heap.c@> +=
static void @
pq_dump (const struct pq *pq) @
{
  int i;

  for (i = 1; i <= pq->n; i++) @
    {@-
      printf ("%d", pq->vertex[i]);
      if (pq->key[i] != DBL_MAX)
	printf (":%g", pq->key[i]);
      putchar (' ');
    }@+
  putchar ('\n');
}

@

@node Binary heap class implementation,  , Binary heap dump, Binary Heap PQ
@subsection Class Implementation

The class implementation of the binary heap is the final declaration in
the file:

@(pq-bin-heap.c@> +=
struct pq_class unf_pq_bin_heap = @
  {@-
    pq_create, @
    pq_discard, @
    pq_extract_min, @
    pq_decrease_key,
    pq_count, @
    pq_key, @
    pq_verify, @
    pq_dump,
  };@+
@

@node Fibonacci Heap PQ,  , Binary Heap PQ, Priority Queues
@section Fibonacci Heap

File @(pq-fib-heap.c@> implements Fibonacci heaps.  It begins this way:

@(pq-fib-heap.c@> =
@<BSD License@>

#include <assert.h>
#include <float.h>
#include <stdio.h>
#include "pq.h"
#include "uniformity.h"

@

A Fibonacci heap is a complicated data structure, quite a bit more
involved to explain than a simple binary heap.  As a result, only the
implementation will be presented here.  For a complete description of
the principles behind a Fibonacci heap, consult one of the references.

The Fibonacci heap priority queue structure is defined this way:

@(pq-fib-heap.c@> +=
/* Fibonacci heap priority queue. */
struct pq @
  {@-
    int n;                      /* Number of items in heap. */
    int m;                      /* Maximum number of items in heap. */

    struct unf_mem *mem;        /* Memory allocator. */
    struct node *min;           /* Minimum node. */
    struct node **index;        /* Index by vertex number to get node.  */
    struct node **rank;         /* Used during extract_min(). */
    struct node *prealloc;      /* Memory allocated for nodes. */
  };@+

@

@noindent
|struct node| has this form:

@(pq-fib-heap.c@> +=
/* Fibonacci heap node. */
struct node @
  {@-
    int vertex;                 /* Vertex value. */
    double key;                 /* Key value. */

    int rank;                   /* Number of direct children. */
    int mark;                   /* 1=marked, 0=unmarked. */

    struct node *parent;        /* Parent node, |NULL| if a root. */
    struct node *child;         /* One of our children, |NULL| if terminal. */
    struct node *next, *prev;   /* Next and previous in circular list. */
  };@+

@

The last bit before the code itself is a set of prototypes for internal
functions:

@(pq-fib-heap.c@> +=
static void insert (struct pq *pq, int vertex, double key);
static void detach (struct node *node);
static void add_root (struct pq *pq, struct node *root);
static void add_child (struct node *parent, struct node *child);

@

@references
@bibref{Cormen 1990}, chapter 21;
@bibref{Fredman 1987};
@bibref{Boyer 1997}.

@menu
* Fibonacci Heap Creation::     
* Fibonacci Heap Destruction::  
* Fibonacci Heap Extract Minimum::  
* Fibonacci Heap Decrease Key::  
* Fibonacci Heap Count::        
* Fibonacci Heap Key Lookup::   
* Fibonacci Heap Verification::  
* Fibonacci Heap Dumping::      
* Fibonacci Heap Class Implementation::  
@end menu

@node Fibonacci Heap Creation, Fibonacci Heap Destruction, Fibonacci Heap PQ, Fibonacci Heap PQ
@subsection Creation

Creating a Fibonacci tree is a relatively simple matter of allocating
memory and initializing all the fields of |struct pq| properly, then
inserting all the nodes with initial keys of @altmath{\infty, +oo}:

@(pq-fib-heap.c@> +=
static struct pq *@
pq_create (struct unf_mem *mem, int n) @
{
  struct pq *pq;
  int i;

  assert (n >= 0);

  pq = mem->unf_alloc (sizeof *pq);
  if (pq == NULL)
    return NULL;

  pq->mem = mem;
  pq->n = 0;
  pq->m = n;
  pq->min = NULL;
  pq->index = pq->mem->unf_alloc (sizeof *pq->index * n);
  pq->rank = pq->mem->unf_alloc (sizeof *pq->rank * n);
  pq->prealloc = pq->mem->unf_alloc (sizeof *pq->prealloc * n);
  if (pq->index == NULL || pq->prealloc == NULL) @
    {@-
      mem->unf_free (pq);
      return NULL;
    }@+

  for (i = 0; i < n; i++) @
    {@-
      pq->index[i] = NULL;
      pq->rank[i] = NULL;
    }@+

  for (i = 0; i < n; i++)
    insert (pq, i, DBL_MAX);

  return pq;
}

@

The |insert()| helper function for |pq_create()| inserts the specified
vertex/key pair into the root list, using the |(n + 1)|th element of
|prealloc| to do it:

@(pq-fib-heap.c@> +=
static void @
insert (struct pq *pq, int vertex, double key) @
{
  struct node *node;

  assert (pq != NULL);
  assert (vertex >= 0 && vertex < pq->m);
  assert (pq->index[vertex] == NULL);

  pq->index[vertex] = node = pq->prealloc + pq->n++;

  node->vertex = vertex;
  node->key = key;
  node->rank = 1;
  node->mark = 0;
  node->parent = node->child = NULL;
  add_root (pq, node);
  if (node->key < pq->min->key)
    pq->min = node;
}

@

@node Fibonacci Heap Destruction, Fibonacci Heap Extract Minimum, Fibonacci Heap Creation, Fibonacci Heap PQ
@subsection Destruction

Destroying a Fibonacci heap is easy:

@(pq-fib-heap.c@> +=
static void @
pq_discard (struct pq *pq) @
{
  assert (pq != NULL);

  pq->mem->unf_free (pq->index);
  pq->mem->unf_free (pq->prealloc);
  pq->mem->unf_free (pq->rank);
  pq->mem->unf_free (pq);
}

@

Notice that because all of the nodes were allocated at once from the
|prealloc| array, there is no need to free each of them separately, nor
would it be correct to do so.

@node Fibonacci Heap Extract Minimum, Fibonacci Heap Decrease Key, Fibonacci Heap Destruction, Fibonacci Heap PQ
@subsection Minimum Value Extraction

Extracting the minimum value from a Fibonacci heap is complicated.
Consult one of the references for details.  Armed with one of those, the
comments in the code below should be sufficient explanation:

@(pq-fib-heap.c@> +=
static int @
pq_extract_min (struct pq *pq) @
{
  struct node *min;
  
  assert (pq != NULL);
  assert (pq->n > 0);

  /* Grab the minimum value and remove it from the heap. */
  min = pq->min;
  pq->index[min->vertex] = NULL;
  pq->n--;
  pq->min = min->next;
  if (pq->min == min)
    pq->min = NULL;
  detach (min);

  /* Add all of the children of the former minimum as roots. */
  for (;;) @
    {@-
      struct node *child = min->child;
      if (child == NULL)
	break;

      detach (child);
      add_root (pq, child);
    }@+

  /* If the heap is now empty, return early. */
  if (pq->min == NULL)
    return min->vertex;

  /* Go through all the root nodes and put them into |rank[]|,
     linking together the ones with equal |rank| fields. */
  {
    struct node *w = pq->min;

    do @
      {@-
	struct node *x = w;
	w = w->next;
	if (w == pq->min)
	  w = NULL;

	for (;;) @
	  {@-
	    struct node *y = pq->rank[x->rank];
	    if (y == NULL || y == x)
	      break;
	    pq->rank[x->rank] = NULL;

	    if (x->key > y->key) @
	      {@-
		struct node *t = x;
		x = y;
		y = t;
	      }@+
	    if (pq->min == y)
	      pq->min = y->next;
	    detach (y);
	    add_child (x, y);
	    y->mark = 0;
	  }@+
	pq->rank[x->rank] = x;
      }@+ @
    while (w != NULL);
  }

  /* Go through the root nodes and unset the corresponding |rank[]| @
     elements.
     Pick the next |pq->min| while we're at it. */
  {
    struct node *x;

    x = pq->min;
    for (;;) @
      {@-
	pq->rank[x->rank] = NULL;
	x = x->next;
	if (x == pq->min)
	  break;

	if (x->key < pq->min->key)
	  pq->min = x;
      }@+
  }

  return min->vertex;
}

@

Function |pq_extract_min()| makes use of three helper functions.  The
first of these is |detach()|, which removes takes a specified node out
of its parent's list of children:

@(pq-fib-heap.c@> +=
/* Removes |node| from the heap, making it parentless. */
static void @
detach (struct node *node) @
{
  assert (node->next != NULL && node->prev != NULL);
  node->next->prev = node->prev;
  node->prev->next = node->next;

  if (node->parent != NULL) @
    {@-
      if (node->parent->child == node) @
	{@-
	  node->parent->child = node->next;
	  if (node->parent->child == node)
	    node->parent->child = NULL;
	}@+

      node->parent->rank--;
      node->parent = NULL;
    }@+

#ifndef NDEBUG
  node->next = node->prev = NULL;
#endif
}

@

The first of these adds a specified node to the priority queue's root
list.  This node must be parentless, meaning that it should earlier have
been detached with |detach()|:

@(pq-fib-heap.c@> +=
/* Adds |root| as a root of |pq|.  
   |pq->min| is changed only if it is |NULL|,
   so the caller is responsible for updating it if it should decrease. */
static void @
add_root (struct pq *pq, struct node *root) @
{
  assert (root->parent == NULL);
  
  if (pq->min == NULL) @
    {@-
      pq->min = root;
      root->prev = root->next = root;
    }@+ @
  else @
    {@-
      root->prev = pq->min;
      root->next = pq->min->next;
      pq->min->next->prev = root;
      pq->min->next = root;
    }@+
}

@

The second helper function adds a node to an arbitrary parent node's
child list.  Again, |child| must be parentless:

@(pq-fib-heap.c@> +=
/* Adds |child| as a child of |parent|. */
static void @
add_child (struct node *parent, struct node *child) @
{
  assert (parent != NULL && child != NULL);
  assert (child->parent == NULL);

  child->parent = parent;
  if (parent->child == NULL) @
    {@-
      child->next = child->prev = child;
      parent->child = child;
    }@+ @
  else @
    {@-
      child->prev = parent->child;
      child->next = parent->child->next;
      parent->child->next->prev = child;
      parent->child->next = child;
    }@+

  parent->rank++;
}

@

@node Fibonacci Heap Decrease Key, Fibonacci Heap Count, Fibonacci Heap Extract Minimum, Fibonacci Heap PQ
@subsection Decreasing Keys

The function for decreasing keys in a Fibonacci tree is not as
complicated as for extracting the minimum key, but it is still
nontrivial, mostly due to the need for ``cascading cuts'' in some
circumstances.  Again, consult the references for full explanation:

@(pq-fib-heap.c@> +=
static void @
pq_decrease_key (struct pq *pq, int vertex, double key) @
{
  struct node *x, *y;
  
  assert (pq != NULL);
  assert (vertex >= 0 && vertex < pq->m);
  assert (pq->index[vertex] != NULL);
  assert (key <= pq->index[vertex]->key);

  /* Find the node in question. */
  x = pq->index[vertex];
  x->key = key;
  y = x->parent;

  /* Make cascading cuts as necessary. */
  if (y != NULL && x->key < y->key)
    for (;;) @
      {@-
	struct node *z;

	assert (x->parent != NULL);
	assert (x->parent == y);
	assert (pq->min != x);
	detach (x);
	add_root (pq, x);
	x->mark = 0;

	z = y->parent;
	if (z == NULL)
	  break;

	if (y->mark == 0) @
	  {@-
	    y->mark = 1;
	    break;
	  }@+ @
	else @
	  {@-
	    x = y;
	    y = z;
	  }@+
      }@+

  /* Actually decrease the key's value. */
  if (key < pq->min->key)
    pq->min = pq->index[vertex];
}

@node Fibonacci Heap Count, Fibonacci Heap Key Lookup, Fibonacci Heap Decrease Key, Fibonacci Heap PQ
@subsection Count

The count is stored in |struct pq|, just as for the binary heap
implementation:

@(pq-fib-heap.c@> +=
static int @
pq_count (const struct pq *pq) @
{
  assert (pq != NULL);
  
  return pq->n;
}

@

@node Fibonacci Heap Key Lookup, Fibonacci Heap Verification, Fibonacci Heap Count, Fibonacci Heap PQ
@subsection Key Lookup

Looking up a key is just a matter of consulting the index:

@(pq-fib-heap.c@> +=
static double @
pq_key (const struct pq *pq, int vertex) @
{
  struct node *node;

  assert (pq != NULL);
  assert (vertex >= 0 && vertex < pq->m);
  
  node = pq->index[vertex];
  return node != NULL ? node->key : -DBL_MAX;
}

@

@node Fibonacci Heap Verification, Fibonacci Heap Dumping, Fibonacci Heap Key Lookup, Fibonacci Heap PQ
@subsection Verification

Extensive verification of the Fibonacci heap's structure can be done:

@(pq-fib-heap.c@> +=
static void @
verify_recursive (const struct pq *pq, const struct node *first, 
                  const struct node *parent, const struct node **tmp) @
{
  const struct node *iter;
  int count;
  
  if (first == NULL)
    return;

  /* Check that the forward links are circular. */
  iter = first;
  count = 0;
  do @
    {@-
      assert (iter->parent == parent);
      assert (count < pq->m);
      tmp[count++] = iter;
      iter = iter->next;
    }@+ @
  while (iter != first);

  /* Check that the backward links are circular,
     and that they are the same ones as when we go forward. */
  iter = first;
  do @
    {@-
      iter = iter->prev;

      assert (count > 0);
      count--;
      assert (tmp[count] == iter);
    }@+ @
  while (iter != first);
  assert (count == 0);

  /* Perform verification recursively on child nodes. */
  iter = first;
  do @
    {@-
      verify_recursive (pq, iter->child, iter, tmp);
      iter = iter->next;
    }@+ @
  while (iter != first);
}

static void @
pq_verify (const struct pq *pq) @
{
  const struct node **tmp;

  assert (pq != NULL);
  assert (pq->n >= 0 && pq->n <= pq->m);
  
  tmp = pq->mem->unf_alloc (sizeof *tmp * pq->n);
  if (pq->n != 0 && tmp == NULL) @
    {@-
      printf ("virtual memory exhausted\n");
      abort ();
    }@+

  verify_recursive (pq, pq->min, NULL, tmp);

  pq->mem->unf_free (tmp);
}

@

@node Fibonacci Heap Dumping, Fibonacci Heap Class Implementation, Fibonacci Heap Verification, Fibonacci Heap PQ
@subsection Dumping

We dump out the full multilevel recursive structure of the Fibonacci
tree:

@(pq-fib-heap.c@> +=
static void @
dump_recursive (const struct node *p) @
{
  const struct node *q = p;

  putchar ('(');
  do @
    {@-
      if (q != p)
	fputs (", ", stdout);
      printf (q->key == DBL_MAX ? "+oo" : "%g", q->key);
      if (q->child != NULL)
	dump_recursive (q->child);

      q = q->next;
    }@+ @
  while (q != p);
  putchar (')');
}

static void @
pq_dump (const struct pq *pq) @
{
  dump_recursive (pq->min);
}

@

@node Fibonacci Heap Class Implementation,  , Fibonacci Heap Dumping, Fibonacci Heap PQ
@subsection Class Implementation

The class implementation of the Fibonacci heap is the final declaration
in the file:

@(pq-fib-heap.c@> +=
struct pq_class unf_pq_fib_heap = @
  {@-
    pq_create, @
    pq_discard, @
    pq_extract_min, @
    pq_decrease_key,
    pq_count, @
    pq_key, @
    pq_verify, @
    pq_dump,
  };@+
@

