@c -*-texinfo-*-

@node Uniformity Testing, Minimum Spanning Tree Computation, Introduction, Top
@chapter Uniformity Testing

The main function of the library presented in this report is to compare
the distribution of points in a multidimensional space against a uniform
distribution.  The library takes as input a set |p| of |n| points in |d|
dimensions each and returns as output a single real number that
signifies how uniformly the points are distributed in |d|-space.

The algorithm used for uniformity testing is based on a minimum spanning
tree.  It can be outlined as follows:

@enumerate 1
@item
Generate a set |q| of |n| points uniformly distributed over the
approximate convex hull of |p|.

@item
Find a minimum spanning tree (MST) for the set of points containing both
|p| and |q|.

@item
Calculate test statistic |t|, based on the MST results, as well as its
expected value |e| and variance |var| given that |p| is uniformly
distributed.

@item
Compute the normal significance |s| of the difference between |t| and
|e|.
@end enumerate

@noindent
The smaller the value of |s| determined in the final step, the more
likely it is that the set |p| is not representative of a uniform
distribution.  Typically a small threshold value such as 0.05 or 0.02 is
compared to |s| in order to make the final decision on uniformity.

This chapter presents the external interface to the uniformity testing
routines as well as their high-level implementation.

@references
@bibref{Smith 1984}, pp.@: 76--77.

@menu
* External Interface::          
* Implementation::              
@end menu

@node External Interface, Implementation, Uniformity Testing, Uniformity Testing
@section External Interface: @file{uniformity.h}

The entire external interface to the uniformity testing routines is in
@file{uniformity.h}.  All externally visible identifiers in this file
and others are prefixed by |unf_| to avoid polluting global namespace.
This is the only header file that needs to be included to use this
library.

Here is the header file's high-level outline:

@(uniformity.h@> =
@<BSD License@>

#ifndef UNIFORMITY_H
#define UNIFORMITY_H

#include <stddef.h>

@<Test options@>
@<High-level uniformity test routine prototype@>
@<Low-level uniformity test routines prototypes@>

#endif /* uniformity.h */
@

@noindent
Most of the time, only one function in this file, |unf_test()|, will be
of interest.  It performs all the steps needed for a MST-based
uniformity test, using a straightforward interface.  The function takes
a set of points |p| and returns a value representing how uniformly the
points are distributed.  It has the following parameters:

@table @asis
@item |struct unf_options *options|
A set of configuration options, which may be set to |NULL| if no special
configuration is desired.

@item |const double *P|
The set of points to test, a total of |n * d| |double|s.  The first |d|
|double|s make up the first point, the second |d| |double|s make up the
second point, and so on.

@item |int n|
The number of points in |p|.

@item |int d|
The number of dimensions in the points in |p|.
@end table

@noindent
|unf_test()| returns a number between 0 and 1, where 0 is very
non-uniform and 1 is completely uniform.

@<High-level uniformity test routine prototype@> =
/* High-level routine. */
double unf_test (const struct unf_options *, const double *, int, int);

@

@menu
* Low-Level Routines::          
* Options::                     
@end menu

@node Low-Level Routines, Options, External Interface, External Interface
@subsection Low-Level Routines

Three low-level routines are also provided for use by external programs:

@table @asis
@item |unf_inside_hull()|
Tests whether a provided point is within the approximate convex hull of
a set of points.

@item |unf_run_mst()|
Finds the MST of a set of points and calculates statistics based on the
results.

@item |unf_calc_results()|
Calculates the significance of test results.
@end table

@noindent
These functions will be discussed in more detail and implemented later
in this chapter.  Here are their prototypes:

@<Low-level uniformity test routines prototypes@> =
/* Low-level routines. */
int unf_inside_hull (const double *, int, int, const double *, double *);
int unf_run_mst (struct unf_mem *, struct unf_mst *, @
		 int *, int *, const double *, int, int);
double unf_calc_results (int, int, int);

@

@node Options,  , Low-Level Routines, External Interface
@subsection Options

Occasionally, it may be desirable to fine-tune the operation of
|unf_test()|.  This possibility is supported by the ``hooks'' provided by
|struct unf_options|, defined as follows:

@<Test options@> =
/* Options. */

/* A set of options for |unf_test()|. */
struct unf_options @
  {@-
    struct unf_mem *unf_mem;    /* Memory allocator. */
    struct unf_rng *unf_rng;    /* Random number generator. */
    struct unf_set *unf_set;    /* Point set. */
    struct unf_mst *unf_mst;    /* Minimum spanning tree. */
  };@+

@

@noindent
To use this structure, declare an instance of it, initialize its members
to their defaults, then set as desired any members that should not take
their default values.  This technique provides for the best
forward-compatibility should a new version of this library be released
with additional options.

Two methods are provided to initialize members of |struct unf_options|
to their defaults.  At declaration time, macro |UNF_DEFAULTS| may be
used as an initializer.  Alternatively, a previously declared option
structure may be initialized with |unf_init_options()|:

@<Test options@> +=
/* Default options. */
#define UNF_DEFAULTS {NULL, NULL, NULL, NULL}
void unf_init_options (struct unf_options *);

@

@noindent
Each of the members of |struct unf_options| points to an object-oriented
``class'' structure.  These classes are used during uniformity testing
for various tasks: memory allocation, random number generation, point
selection, and minimum spanning tree calculation.  The defaults are
generally reasonable, but these hooks are provided to allow for
substitution of better alternatives.  Later chapters will discuss these
classes in detail.  For now, it is sufficient to list the
implementations provided here:

@<Test options@> +=
/* Provided class implementations. */
extern struct unf_mem unf_mem_malloc;
extern struct unf_set unf_set_rectangular;
extern struct unf_rng unf_rng_mt;
extern struct unf_rng unf_rng_system;
extern struct unf_mst unf_mst_prim_binary;
extern struct unf_mst unf_mst_prim_fibonacci;

@

@node Implementation,  , External Interface, Uniformity Testing
@section Implementation: @file{uniformity.c}

The functions declared in @file{uniformity.h} are implemented in
@file{uniformity.c}, which begins as follows:

@(uniformity.c@> =
@<BSD License@>

#include <assert.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "uniformity.h"

@

@menu
* unf_test() implementation::   
* unf_inside_hull() implementation::  
* unf_run_mst() implementation::  
* unf_calc_results() implementation::  
* unf_init_options() implementation::  
@end menu

@node unf_test() implementation, unf_inside_hull() implementation, Implementation, Implementation
@subsection High-Level Test Function

Here is the outline of the unf_test() function:

@(uniformity.c@> +=
/* Tests for uniformity of |p|, a set of |n| points in |d| dimensions.
   |p| is an array of |n * d| doubles, where the first point is stored
   in the first |d| elements, the second in the next |d| elements, and
   so on.
   The contents in |user_options|, if non-|NULL|, are used to control
   the testing procedure.
   Returns a confidence level between 0 and 1, 1 indicating that the
   set is perfectly uniform and 0 indicating that the set is perfectly
   clustered.  Returns |-1| in the event of a memory allocation
   error. */
double @
unf_test (const struct unf_options *user_options, @
	  const double *p, int n, int d) @
{
  struct unf_options options;   /* Local copy of options. */
  double *r;                    /* Stores |p| plus generated points. */

  assert (p != NULL && n > 0 && d > 0);

  @<Set up |options| for uniformity test@>
  @<Allocate memory for uniformity test@>
  @<Generate points for uniformity test@>
  @<Run MST, calculate results, and return@>
}

@

The first step is to set up the |user_options| argument.  If it is
|NULL|, we need to replace it entirely by the defaults.  Otherwise, we
make a copy and check member by member for nulls:

@<Set up |options| for uniformity test@> =
{
  static const struct unf_options default_options = @
    {@-
      &unf_mem_malloc, @
      &unf_rng_system, @
      &unf_set_rectangular, @
      &unf_mst_prim_binary,
    };@+

  if (user_options == NULL) @
    options = default_options;
  else @
    {@-
      options = *user_options;
      if (options.unf_mem == NULL) @
        options.unf_mem = default_options.unf_mem;
      if (options.unf_rng == NULL) @
        options.unf_rng = default_options.unf_rng;
      if (options.unf_set == NULL) @
        options.unf_set = default_options.unf_set;
      if (options.unf_mst == NULL) @
        options.unf_mst = default_options.unf_mst;
    }@+
}

@

The second step is to allocate memory, using the memory allocator in
|options.unf_mem|.  We need storage for |(2 * n)| points of |d|
dimensions for storing the points from |p| and an equal number of
generated points, plus space for 2 more such points for temporary
storage.  In order to save on calls for memory allocation, we
consolidate all this temporary storage into a single block and allocate
it all at once.

@<Allocate memory for uniformity test@> =
r = options.unf_mem->unf_alloc (sizeof *r * ((2 + n * 2) * d));
if (r == NULL)
  return -1.0;

@

The third step is to generate the points in |r| on which to perform the
MST.  The first |n| of these points are copied from |p|.  It would be
faster here to avoid the copy, passing instead a pair of arrays to the
MST calculation routine, but it would both complicate and slow down the
MST routine itself, so such a change would probably not be worthwhile.

The remaining |n| of the points are generated randomly within the
approximate convex hull of |p|.  This is done using a rejection
procedure: points are picked randomly from a conveniently shaped volume
that contains the convex hull (for instance, a hyper-rectangle or
hypersphere), then points that do not fall within the approximate
convex hull are discarded.

The routine for testing whether a point falls within the approximate
convex hull of |p|, |unf_inside_hull()|, uses the storage allocated
earlier for an additional pair of points.  Notice that this is one place
where being able to use C99 features would be handy.  Variable Length
Arrays (VLAs) from C99 would allow these additional points to be
declared as local variables with no need to pre-allocate and pass them
down, improving modularity.

@<Generate points for uniformity test@> =
{
  double *const tmp = r + (n * 2) * d;
  struct unf_a_set *s;
  int i;

  s = options.unf_set->unf_create (options.unf_mem, p, n, d);
  if (s == NULL) @
    {@-
      options.unf_mem->unf_free (r);
      return -1.0;
    }@+

  memcpy (r, p, sizeof *r * n * d);
  for (i = 0; i < n; ) @
    {@-
      double *y = r + (n + i) * d;
      options.unf_set->unf_random (s, options.unf_rng, y);
      if (unf_inside_hull (p, n, d, y, tmp))
        i++;
    }@+

  options.unf_set->unf_discard (s);
}

@

Finally, we run the MST and calculate and return the results:

@<Run MST, calculate results, and return@> =
{
  int c, t;
  int okay;

  okay = unf_run_mst (options.unf_mem, options.unf_mst, &c, &t, r, n, d);
  options.unf_mem->unf_free (r);
  return okay ? unf_calc_results (c, t, n) : -1;
}
@

@node unf_inside_hull() implementation, unf_run_mst() implementation, unf_test() implementation, Implementation
@subsection Convex Hull Insideness Testing Function

Function |unf_inside_hull()| determines whether a given point |y| is
within the approximate convex hull of a set of points |p|, using the
algorithm specified.  This algorithm first calculates the vector
@tex
$$\widehat{n}^* = {1\over n} \sum_{i=1}^n{{p[i] - y}\over{(\Vert p[i] -
y\Vert^2)^{d+1}}}\ ,$$
@end tex
@ifnottex
@example
                   __
                   \
n_star_est = (1/n)  > (p[i] - y) * length(p[i] - y))**(d + 1)
                   /_
                i=1...n
@end example

@end ifnottex
@noindent
where, as usual, |n| is the number of points and |d| the number of
dimensions.  This vector is an estimate of @altmath{n^*, |n_star|}, a
vector for which the dot product @altmath{n^* \cdot (p[i] - y) > 0,
n_star * (p[i] - y) > 0} is true for all @w{|i| =} 1, 2, @dots{}, |n|,
which is true if and only if @altmath{n^*, |n_star|} is within the
convex hull of |p|.  The test calculates this dot product for all values
of |i| and reports that |y| is within |p|'s convex hull only if the
result is always positive.

The first part of the implementation is a helper function to calculate
@altmath{\widehat{n}^*, |n_star_est|}.  This function,
|calc_n_star_est()|, iterates over the values of |i|.  For each |i|, it
measures the magnitude of the vector difference |p[i] - y| and
calculates a scalar |factor| representing the denominator in the
equation above.  The vector difference, times |factor|, is added to the
running sum for @altmath{\widehat{n}^*, |n_star_est|}.  Finally, the
@math{1/n} factor is applied, and the function returns.

@(uniformity.c@> +=
/* Calculates vector |n_star_est| at point |y| within points |p|.
   |tmp| is used for temporary storage.  Its contents are destroyed.
   There are |d| elements in each of |n_star|, |tmp|, and |y|, and |n * @
   d| elements in |p|. */
static void @
calc_n_star_est (double *n_star_est, double *tmp,
                 const double *p, int n, int d, const double *y) @
{
  int i;

  assert (n_star_est != NULL && tmp != NULL && p != NULL
	  && n > 0 && d > 0 && y != NULL);

  for (i = 0; i < d; i++)
    n_star_est[i] = 0.0;

  for (i = 0; i < n; i++) @
    {@-
      double length = 0.0;
      double factor;
      int j;

      for (j = 0; j < d; j++) @
	{@-
	  double diff = tmp[j] = p[j] - y[j];
	  length += diff * diff;
	}@+
      factor = 1.0 / pow (length, d + 1.0);

      for (j = 0; j < d; j++)
	n_star_est[j] += tmp[j] * factor;

      p += d;
    }@+

  for (i = 0; i < d; i++)
    n_star_est[i] /= n;
}

@

The |unf_inside_hull()| function itself is simple.  It calls
|calc_n_star_est()| to calculate |n_star_est|, then calculates the dot
product |dot| of |n_star_est| with the vector difference |p[i] - y| for
each possible value of |i|.  If |dot| is positive every time, then |y|
is within the approximate convex hull of |p|; otherwise, it is not.

@(uniformity.c@> +=
/* Returns nonzero only if point |y| is within the (approximate) convex
   hull of the |n| points in array |p|.
   Each point has |d| dimensions, and |tmp| must point to a modifiable
   scratch array with room for |2 * d| doubles. */
int @
unf_inside_hull (const double *p, int n, int d, const double *y, @
		 double *tmp) @
{
  double *n_star_est;
  int i;

  assert (p != NULL && n > 0 && d > 0 && y != NULL && tmp != NULL);

  n_star_est = tmp + d;
  calc_n_star_est (n_star_est, tmp, p, n, d, y);

  for (i = 0; i < n; i++) @
    {@-
      double dot;
      int j;

      dot = 0.0;
      for (j = 0; j < d; j++)
	dot += (*p++ - y[j]) * n_star_est[j];

      if (dot <= 0.0)
	return 1;
    }@+

  return 0;
}
@

@references
@bibref{Smith 1984}, pp.@: 75--76.

@node unf_run_mst() implementation, unf_calc_results() implementation, unf_inside_hull() implementation, Implementation
@subsection MST Test Run Function

The |unf_run_mst()| function actually runs the MST of the pooled
original and generated points, then generates the test statistics |c|
and |t| based on the results.  It works like this:

@(uniformity.c@> +=
/* Runs the MST-based test for uniformity on the |2 * n| points in @
   |r|, of |d| dimensions each.
   The first |n| of the points should be the points to test, the last
   |n| the randomly generated points within the approximate convex
   hull of those points.
   |*c| and |*t| receive on output the C and T statistics.
   Returns nonzero if successful, zero if a memory allocation error @
   occurred. */
int @
unf_run_mst (struct unf_mem *mem_class, struct unf_mst *mst_class,
	     int *c, int *t, const double *r, int n, int d) @
{
  unf_mst_result *mst;
  int i;

  assert (mem_class != NULL && mst_class != NULL && c != NULL && t != NULL
	  && r != NULL && n > 0 && d > 0);

  @<Find MST of pooled points@>
  @<Calculate T statistic@>
  @<Calculate C statistic@>
  @<Free memory and return@>
}

@

We delegate the computation of the MST to a hook function.  The function
takes a memory allocator class and the set of points as its arguments.
It returns a pointer to |2 * n - 1| pairs of |int|s, where each pair
designates the indexes of two elements of |r| connected within the MST.

@<Find MST of pooled points@> =
mst = mst_class->unf_run (mem_class, r, 2 * n, d);
if (mst == NULL)
  return 0;

@

The T statistic is the number of edges in the MST that join a point in
the original set of points |p| with a point in the set of randomly
generated points.  The original points are the first |n| in |r|, the
generated ones are the remainder, so this is simple.

@<Calculate T statistic@> =
*t = 0;
for (i = 0; i < 2 * n - 1; i++)
  if ((mst[i][0] < n) != (mst[i][1] < n))
    (*t)++;

@

The C statistic is the number of pairs of edges in the MST that share a
node in common.

@<Calculate C statistic@> =
*c = 0;
for (i = 0; i < 2 * n - 1; i++) @
  {@-
    int j;

    for (j = i + 1; j < 2 * n - 1; j++)
      if (mst[i][0] == mst[j][0] @
          || mst[i][0] == mst[j][1]
          || mst[i][1] == mst[j][0] @
          || mst[i][1] == mst[j][1])
        (*c)++;
  }@+

@

Finally, we finish up by freeing the memory allocated for the MST and
returning a successful value.

@<Free memory and return@> =
mem_class->unf_free (mst);
return 1;
@

@node unf_calc_results() implementation, unf_init_options() implementation, unf_run_mst() implementation, Implementation
@subsection Results Calculation Function

The final step in uniformity testing is to analyze the results based on
the MST from the previous step, using |c| and |t| computed there.
Qualitatively, if the test data are uniformly distributed, then we
expect there to be about as many edges joining points within the two
groups of points (the test set and the generated set) as edges between
the two groups.  But if the test data are clustered, then we expect most
edges to join points within a group, with a few edges joining the two
groups.

This notion is quantifiable.  The expected or mean value @math{E[t]} of
|t|, given that the data are uniformly distributed, is |n|, the number
of test points, with variance
@tex
$${\rm Var}[t\vert c] =
{n\over 2n - 1}
\left( n - 1 - {c - 2n + 2\over 2n - 3} \right),
$$
@end tex
@ifnottex
@example
                               _                   _
                          n   !         c - 2n + 2  !
            Var[t!c] = ------ ! n - 1 - ----------  ! ,
                       2n - 1 !_          2n - 3   _!
@end example

@end ifnottex
@noindent
and the distribution of |t| conditioned on |c| is asymptotically normal
in |n|.  We treat |t| as a normal variate and convert it to a |z|-score
using the formula
@tex
$$z = {t - E[t]\over \sqrt{{\rm Var}[t\vert c]}}\,.
$$
@end tex
@ifnottex
@example
z = (t - E[t]) / sqrt(Var[t!c])
@end example

@end ifnottex
@noindent
The significance |s| is then found using the cumulative distribution
function @math{P(z)} of the standard normal distribution @math{Z(z)},
@tex
$$s = P(z) = \int\limits_{-\infty}^z Z(x)\ dx,\quad
Z(z) = {1\over \sqrt{2\pi}}\,e^{-x^2/2}.
$$
@end tex
@ifnottex
@example
           / z
s = P(z) = !     Z(x) dx,  Z(z) = 1 / (2 * pi) * exp(-0.5 * x**2).
           / -oo
@end example

@end ifnottex
@noindent
The significance is returned to the caller, which can then apply its own
criteria to make the final decision about uniformity.  Typically, the
caller classifies as uniform samples with @altmath{s > s_0, s > s_0} and
as non-uniform samples with @altmath{s \le s_0, s <= s_0}, where
@altmath{s_0, s_0} is a small threshold value such as 0.02 or 0.05.

@blankline

The implementation is straightforward.  To find the results, calculate
the mean and variance of |t| conditioned on |c|, then report the
significance of the resulting |z|-score:

@(uniformity.c@> +=
@<Significance of a normal variate@>

/* Returns uniformity significance level between 0 and 1.
   |c| and |t| are the values returned by |unf_run_mst()|, |n| the @
   size of the test set. */
double @
unf_calc_results (int c, int t, int n) @
{
  double mean = n;
  double var = ((n / (2.0 * n - 1.0))
		* (n - 1.0 - (c - 2.0 * n + 2.0) / (2.0 * n - 3.0)));
  double z = (t - mean) / sqrt (var);
  
  return normal_sig (z);
}

@

The code above makes use of a routine |normal_sig()| to compute the
significance.  The significance P(z) of a standard normal variate |z @>=
0| can be calculated using the formula
@tex
$$P(z) = 1 - Z(z)[a_1t + a_2t^2 + a_3t^3] + \epsilon(x),\quad 
t = 1/(1 + px)
$$
@end tex
@ifnottex
P(z) = 1 - Z(z)[(a_1)t + (a_2)t**2 + (a_3)t**3] + epsilon(x),  t = 1/(1 + px)
@example
@end example

@end ifnottex
@noindent
where @altmath{a_1, a_1}, @altmath{a_2, a_2}, and @altmath{a_3, a_3} are
constants and @altmath{\epsilon(x) < 10^{-5}, e(x) < .00001} is an error
term.  For |z < 0|, we can use the identity @math{P(x) = 1 - P(-x)}.
The implementation in C is as follows:

@<Significance of a normal variate@> =
/* Returns the significance of normal variate |x|. */
static double @
normal_sig (double x) @
{
  const double a1 = 0.4361836;
  const double a2 = -0.1201676;
  const double a3 = 0.9372980;
  const double p = 0.33267;
  const double pi = 3.14159265358979323846;

  double y = fabs (x);
  double t = 1.0 / (1.0 + p * y);
  double t2 = t * t;
  double t3 = t2 * t;
  double z = 1.0 / sqrt (2.0 * pi) / exp (0.5 * y * y);
  double s = z * (a1 * t + a2 * t2 + a3 * t3);

  return x >= 0 ? 1.0 - s : s;
}
@

@references
@bibref{Smith 1984}, pp.@: 76;
@bibref{Abramowitz 1972}, pp.@: 931--932;
@bibref{Zwillinger 1996}, pp.@: 583--584. 

@node unf_init_options() implementation,  , unf_calc_results() implementation, Implementation
@subsection Options Initialization Function

@(uniformity.c@> +=
/* Initializes |options| to the defaults, as an alternative to using
   |unf_defaults| as an initializer. */
void @
unf_init_options (struct unf_options *options) @
{
  assert (options != NULL);
  
  options->unf_mem = NULL;
  options->unf_rng = NULL;
  options->unf_set = NULL;
  options->unf_mst = NULL;
}
@
